"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.HttpCheckService = void 0;
const logger = require("../../../log");
const version = require("../../../version");
const axios_1 = require("../../../axios");
class HttpCheckService {
    constructor(checkStore) {
        this.checkStore = checkStore;
    }
    async run(checkId) {
        logger.debug({ checkId }, 'executing http check');
        const check = await this.checkStore.get(checkId);
        if (check === null) {
            throw new Error(`Not found check with checkId: '${checkId}'`);
        }
        try {
            const response = await axios_1.axiosInstance.get(check.url, {
                headers: {
                    'User-Agent': `broker client/${version} (http check service)`,
                },
                timeout: check.timeoutMs,
                validateStatus: () => true,
            });
            logger.trace({ response: response.data }, 'http check raw response');
            const checkResult = convertResponseToCheckResult(check, response.status, response.statusText, JSON.stringify(response.data));
            logger.debug({ checkId, status: checkResult.status }, `completed http check execution`);
            return Promise.resolve(checkResult);
        }
        catch (error) {
            const errorMessage = `Error executing check with checkId ${checkId}`;
            logger.debug({ error }, errorMessage);
            throw new Error(errorMessage);
        }
    }
}
exports.HttpCheckService = HttpCheckService;
function convertResponseToCheckResult(check, statusCode, statusText, response) {
    let status;
    if (statusCode >= 200 && statusCode <= 299) {
        status = 'passing';
    }
    else if (statusCode == 429) {
        status = 'warning';
    }
    else {
        status = 'error';
    }
    const output = `HTTP GET ${check.url}: ${statusCode} ${statusText}, response: ${truncate(response)}`;
    return {
        id: check.checkId,
        name: check.checkName,
        output: output,
        status: status,
    };
}
function truncate(response) {
    if (response.length <= 200) {
        return response;
    }
    const truncatedResponse = response.substring(0, 200);
    return `${truncatedResponse}... (truncated)`;
}
//# sourceMappingURL=http-check-service.js.map