"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.checksConfig = exports.executePreflightChecks = exports.preflightChecksEnabled = void 0;
const logger = require("../../log");
const http_check_service_1 = require("./http/http-check-service");
const preflight_check_store_1 = require("./preflight-check-store");
const http_checks_1 = require("./http/http-checks");
const exponential_backoff_1 = require("../retry/exponential-backoff");
function preflightChecksEnabled(config) {
    // preflight checks are enabled per default
    let preflightChecksEnabled = true;
    const preflightChecksEnabledValue = config
        .PREFLIGHT_CHECKS_ENABLED;
    if (typeof preflightChecksEnabledValue !== 'undefined') {
        preflightChecksEnabled =
            preflightChecksEnabledValue.toLowerCase() === 'true' ||
                preflightChecksEnabledValue.toLowerCase() === 'yes';
    }
    logger.info({ enabled: preflightChecksEnabled }, 'verifying if preflight checks are enabled');
    return preflightChecksEnabled;
}
exports.preflightChecksEnabled = preflightChecksEnabled;
async function executePreflightChecks(config) {
    const preflightCheckResults = [];
    const { preflightCheckStore, httpCheckService } = await (0, exports.checksConfig)(config);
    const checks = await preflightCheckStore.getAll();
    for (const check of checks) {
        const checkResult = await (0, exponential_backoff_1.retry)(() => httpCheckService.run(check.checkId), {
            retries: 3,
            operation: `http check ${check.checkId}`,
        });
        preflightCheckResults.push(checkResult);
    }
    logPreflightCheckResults(preflightCheckResults);
    return Promise.resolve(preflightCheckResults);
}
exports.executePreflightChecks = executePreflightChecks;
const configurePreflightCheckStore = async (config) => {
    const preflightCheckStore = new preflight_check_store_1.PreflightCheckStore();
    await preflightCheckStore.add((0, http_checks_1.createBrokerServerHealthcheck)(config));
    await preflightCheckStore.add((0, http_checks_1.createRestApiHealthcheck)(config));
    return Promise.resolve(preflightCheckStore);
};
const configureHttpCheckService = async (checkStore) => {
    const httpCheckService = new http_check_service_1.HttpCheckService(checkStore);
    return Promise.resolve(httpCheckService);
};
// wire all dependencies (services and stores) here
const checksConfig = async (config) => {
    const preflightCheckStore = await configurePreflightCheckStore(config);
    const httpCheckService = await configureHttpCheckService(preflightCheckStore);
    return {
        preflightCheckStore,
        httpCheckService,
    };
};
exports.checksConfig = checksConfig;
const logPreflightCheckResults = (results) => {
    console.log('##############################################################');
    console.log('### PREFLIGHT CHECKS RESULTS');
    console.log('###');
    console.log('### Preflight checks help to catch errors early, upon broker');
    console.log('### client startup. Note, that broker client will start');
    console.log('### whether the checks were successful or not.');
    console.log('###');
    console.log('### See more: https://github.com/snyk/broker#preflight-checks');
    console.log('##############################################################');
    const checks = results.map((check) => {
        return { id: check.id, status: check.status };
    });
    console.table(checks);
};
//# sourceMappingURL=index.js.map